import numpy as np
import pkg_resources
import re

BASE_BACKUP = 'DEFAULT'

VALID_RESPONSE_CODES = [200, 201, 202, 204]

#Base API URL
API_URL = "https://api.3doptix.com/v1"

#GET endpoints
GET_SETUPS_ENDPOINT = 'setups'
GET_SETUP_ENDPOINT = 'setups/{setup_id}'
GET_PART_ENDPOINT = 'setups/{setup_id}/parts/{part_id}'
GET_ANSWER_ENDPOINT = 'http://chatbot.5doptix.com/ask'
MAX_HISTORY_LEN = 5

# DELETE endpoints
DELETE_PART_ENDPOINT = 'setups/{setup_id}/parts/{part_id}'

#PUT endpoints
PUT_BATCH_CHANGES_ENDPOINT = "setups/{setup_id}/batch_changes"
PUT_ANALYSIS_ENDPOINT = 'setups/{setup_id}/run_analyses'
PUT_SIMULATION_ENDPOINT = 'setups/{setup_id}/simulation'

#SET endpoints
SET_PART_ENDPOINT = 'setups/{setup_id}/parts/{part_id}'

#POST endpoints
POST_ADD_ANALYSIS_ENDPOINT = "setups/{setup_id}/parts/{part_id}/add_analyses"
POST_CREATE_OPTICS_ENDPOINT = "create_optics"
POST_CREATE_SETUP_ENDPOINT = 'setups/create'
POST_ADD_PART_ENDPOINT = 'setups/{setup_id}/add_part'

#Snellius version: must be compatible
SNELLIUS_VERSION = "2.2.23"
GPU_TYPE = "g4"

#Messages and warnings
SET_API_URL_WARNING = "Are you sure you want to change the API URL? This is internal option and should not be used by users."
SERVER_DOWN_MESSAGE = "The server is down. Please try again later."
VERSION = pkg_resources.get_distribution('threed_optix').version
WELCOME_MESSAGE = f"Welcome to 3DOptix! 🌈 You are now being connected to the 3DOptix API (version {VERSION}). Let's start!"
GETTING_PARTS_MESSAGE = f"We are retrieving the setup information. This needs to be done only once. Please wait."

# Setups
SETUP_LABELS = ["General",
                "Microscopy",
                "Telescopy" ,
                "Spectroscopy" ,
                "Imaging" ,
                "Non-linear optics" ,
                "Fiber" ,
                "Illumination" ,
                "Light sources" ,
                "Laser Optics" ,
                "Diffractive Optics"
]

INCH = 'in'
MM = 'mm'
SETUP_UNITS = [MM, INCH]
CREATE_SETUP_UNITS_MAP = {
    INCH: 0,
    MM: 1,
}
OPTICS_ADD_PART_TYPE = 2
LIGHT_SOURCE_ADD_PART_TYPE = 8
DETECTOR_ADD_PART_TYPE = 9


# Analysis
## Analysis file decoding format
ANALYSIS_HEADER_DTYPES = np.dtype([
        ('magic number', np.int32),
        ('analysis kind', np.int32),
        ('data kind', np.int32),
        ('polarization kind', np.int32),
        ('version', np.int32),
        ('num_hits', np.int64),
        ('num_wavelengths', np.int32),
        ('resolution_x', np.int32),
        ('resolution_y', np.int32),
    ])
# ANALYSIS_MATRIX_DTYPES = np.dtype(np.float32).newbyteorder('<')
ANALYSIS_MATRIX_DTYPES = np.float32
HEADER_BYTES = 40
DATA_KINDS_MAPPING = {
    -1: 'INVALID_SPOT_DATA_KIND',
    0: 'SPOT_INCOHERENT_IRRADIANCE_S',
    1: 'SPOT_INCOHERENT_IRRADIANCE_P',
    2: 'SPOT_COHERENT_IRRADIANCE_S',
    3: 'SPOT_COHERENT_IRRADIANCE_P',
    4: 'SPOT_COHERENT_PHASE_S',
    5: 'SPOT_COHERENT_PHASE_P',
    6: 'SPOT_INCOHERENT_IRRADIANCE',
    7: 'SPOT_COHERENT_IRRADIANCE',
    8: 'SPOT_COHERENT_PHASE',
    9: 'SPOT_FIELD',
    10: 'SPOT_MTF',
    11: 'SPOT_WAVEFRONT'
}

SUPPORTED_DATA_KINDS_NAMES = [
    'SPOT_INCOHERENT_IRRADIANCE_S',
    'SPOT_INCOHERENT_IRRADIANCE_P',
    'SPOT_COHERENT_IRRADIANCE_S',
    'SPOT_COHERENT_IRRADIANCE_P',
    'SPOT_COHERENT_PHASE_S',
    'SPOT_COHERENT_PHASE_P',
    'SPOT_INCOHERENT_IRRADIANCE',
    'SPOT_COHERENT_IRRADIANCE',
    'SPOT_COHERENT_PHASE',
]

POLARIZATION_MAPPING = {
    0: 'NONE_POLARIZATION',
    1: 'X_POLARIZATION',
    2: 'Y_POLARIZATION',
    3: 'Z_POLARIZATION',
}

## Valid names
FAST_ANALYSIS_NAMES = [
    "Spot (Incoherent Irradiance)",
    "Spot (Incoherent Irradiance) Through Focus",
    "Grid Distortion",
    "Distortion & Field Curvature",
    "OPD (Optical Path Difference)",
    "Ray Abberations (TRA and LRA)",
    "Polarization Map"
]
ADVANCED_ANALYSIS_NAMES = [
    "Spot (Incoherent Irradiance)",
    "Spot (Coherent Irradiance)",
    "Coherent Phase",
    "Spot (Coherent Irradiance) Huygens",
    "Coherent Phase Huygens",
    "Spot (Coherent Irradiance) Fresnel",
    "Coherent Phase Fresnel",
    "Spot (Coherent Irradiance) Polarized",
    "Coherent Phase Polarized",
    "Spot (Incoherent Irradiance) Polarized"
]
ANALYSIS_NAMES =  [
    "Spot (Incoherent Irradiance)",
    "Spot (Coherent Irradiance)",
    "Coherent Phase",
    "Spot (Coherent Irradiance) Huygens",
    "Coherent Phase Huygens",
    "Spot (Coherent Irradiance) Polarized",
    "Coherent Phase Polarized",
    "Spot (Incoherent Irradiance) Polarized"
]

## results color scale (as in GUI)
COLOR_SCALE = [
    [0.0, '#0000FF'],
    [0.25, '#00FF00'],
    [0.5, 'yellow'],
    [0.75, 'orange'],
    [1.0, '#FF0000']
]

# Errors
## simulation errors
SIMULATION_ERROR = """Simulation failed (server side).\n Error message: "{message}"."""

## Analyses errors
ANALYSES_ADD_ERROR = "Analyses {not_added} were not added. Please add them first"
ANALYSES_NOT_SAME_SETUP_ERROR = f"Analyses must be from the same setup."
ANALYSIS_RUN_ERROR = """Analyses failed (server side).\n Error message: "{message}"."""
ANALYSES_DUPLICATED_ERROR = """Analyses with ids {duplicated} are duplicated.\nYou can use the existing analyses with the same parameters or force this action by setting 'force' argument to True"""
ANALYSES_RUN_ERROR = """Analysis failed (server side).\n Error message: "{message}"."""

#Arguments
def argument_repair_message(errors):
    return "Invalid arguments. Errors:\n" + "\n".join(errors)

# Analysis
ANALYSIS_RES_RANGE = [0, 10000]
ANALYSIS_RAYS_RANGE = [1, 1e9]

# Detector Data
## size
DETECTOR_SIZE_RANGE = [0.0001, 200]
DETECTOR_OPACITY_RANGE = [0, 1]
#Light source data
##Beam kinds
GAUSSIAN_BEAM = "GAUSSIAN_BEAM"
POINT_SOURCE = "POINT_SOURCE"
PLANE_WAVE = "PLANE_WAVE"
BEAM_KINDS = [POINT_SOURCE, GAUSSIAN_BEAM, PLANE_WAVE]

## Density patterns
XY_GRID = "XY_GRID"
CONCENTRIC_CIRCLES = "CONCENTRIC_CIRCLES"
RANDOM = "RANDOM"
DENSITY_PATTERNS = [XY_GRID, CONCENTRIC_CIRCLES, RANDOM]

## Count type
TOTAL = "TOTAL"
PER_WAVELENGTH = "PER_WAVELENGTH"
LOWER_VIS_COUNT_LIMIT = 1
UPPER_VIS_COUNT_LIMIT = 200
COUNT_TYPES = [TOTAL, PER_WAVELENGTH]

## Plane wave
RECTANGULAR = "RECTANGULAR"
CIRCULAR = "CIRCULAR"
ELLIPTICAL = "ELLIPTICAL"
PLANE_WAVE_TYPES = [RECTANGULAR, CIRCULAR, ELLIPTICAL]

## Appearance
HEX_COLOR_RE = r'#?([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})'
LIGHT_SOURCE_OPACITY_RANGE = [0, 1]
MODEL_RADIUS_RANGE = [1, 10]
## Angular Information
AZIMUTH_Z_RANGE = [-90, 90]
THETA_RANGE = [-180, 180]
PHI_RANGE = [0,180]

## Power
POWER_RANGE = [0, 1e6]

## Point Source
HALF_CONE_ANGLE = "HALF_CONE_ANGLE"
HALF_WIDTH_RECT = "HALF_WIDTH_RECT"
HALF_WIDTH_AT_Z = "HALF_WIDTH_AT_Z"
POINT_SOURCE_TYPES = [HALF_CONE_ANGLE, HALF_WIDTH_RECT, HALF_WIDTH_AT_Z]
PLANE_WAVE_WIDTH_RANGE = [0.0001, 1e5]
PLANE_WAVE_HEIGHT_RANGE = [0.0001, 1e5]
PLANE_WAVE_RADIUS_RANGE = [0.0001, 1e2]
POINT_SOURCE_ANGLE_RANGE = [0.001, 180]
POINT_SOURCE_DIST_Z_RANGE = [0.0001, 9007199254740991]
POINT_SOURCE_DIST_Z_AT_RANGE = [0.0001, 9007199254740991]


## Wavelegnth
WAVELENGTH_RANGE = [140, 20000]
WEIGHTS_RANGE = [1e-5, 1]

# Create Optics
METALLIC_MIRROR = "Metallic Mirror"
DIELECTRIC_MIRROR = "Dielectric Mirror"
DICHROIC_MIRROR = "Dichroic Mirror"
CURVED_MIRROR_CONCAVE = "Curved Mirror (Concave)"
CURVED_MIRROR_CONVEX = "Curved Mirror (Convex)"
CYLINDRICAL_MIRROR_PLANO_CONVEX = "Cylindrical Mirror (Plano-Convex)"
CYLINDRICAL_MIRROR_PLANO_CONCAVE = "Cylindrical Mirror (Plano-Concave)"
PARABOLIC_MIRROR_CONCAVE = "Parabolic Mirror (Concave)"
OFF_AXIS_PARABOLIC_MIRROR = "Off-Axis Parabolic Mirror"
D_SHAPED_MIRROR = "D-Shaped Mirror"
KNIFE_EDGE_PRISM_MIRROR = "Knife-Edge Prism Mirror"
RIGHT_ANGLE_PRISM_MIRROR = "Right-Angle Prism Mirror"
ROOF_PRISM_MIRROR = "Roof Prism Mirror"
ELLIPSOIDAL = "Ellipsoidal"
KNIFE_EDGE_PRISM = "Knife-Edge Prism"
RIGHT_ANGLE_PRISM = "Right-Angle Prism"
EQUIANGULAR_PRISM = "Equilateral Prism"
DOVE_PRISM = "Dove Prism"
WEDGE_PRISM = "Wedge Prism"
PELLIN_BROCA_PRISM = "Pellin Broca Prism"
PENTA_PRISM = "Penta Prism"
AXICON_PRISM = "Axicon Prism"
RHOMBIC_PRISM = "Rhombic Prism"
MULTIPLETS = "Multiplets"
SPHERICAL_LENS_PLANO_CONVEX = "Spherical Lens (Plano-Convex)"
SPHERICAL_LENS_PLANO_CONCAVE = "Spherical Lens (Plano-Concave)"
SPHERICAL_LENS_CONVEX_CONVEX = "Spherical Lens (Convex-Convex)"
SPHERICAL_LENS_CONCAVE_CONCAVE = "Spherical Lens (Concave-Concave)"
SPHERICAL_LENS_CONCAVE_CONVEX = "Spherical Lens (Concave-Convex)"
SPHERICAL_LENS_CONVEX_CONCAVE = "Spherical Lens (Convex-Concave)"
CYLINDRICAL_LENS_PLANO_CONVEX = "Cylindrical Lens (Plano-Convex)"
CYLINDRICAL_LENS_PLANO_CONCAVE = "Cylindrical Lens (Plano-Concave)"
BALL_LENS = "Ball Lens"
HALF_BALL_LENS = "Half Ball Lens"
OFF_AXIS_PARABOLIC_LENS = "Off-Axis Parabolic Lens"
ASPHERIC_LENS = "Aspheric Lens"
ODD_ASPHERE = "Odd Asphere"
PARAXIAL_LENS = "Paraxial lens"
GENERAL_CONIC = "General Conic"
GENERAL_BICONIC = "General Biconic"
LONGPASS_FILTER = "Longpass Filter"
SHORTPASS_FILTER = "Shortpass Filter"
BANDPASS_FILTER = "Bandpass Filter"
DICHROIC_FILTER = "Dichroic Filter"
NOTCH_FILTER = "Notch Filter"
CALIBRATION_FILTER = "Calibration Filter"
SPECTRAL_SHAPE_FILTER = "Spectral Shape Filter"
LASER_LINE_FILTER = "Laser-line Filter"
COLOR_GLASS = "Color Glass"
NEUTRAL_DENSITY_FILTER = "Neutral Density Filter"
HOT_COLD_MIRROR = "Hot/Cold Mirror"
THIN_BS = "Thin BS"
CUBE_BS = "Cube BS"
CUBE_BS_POLARIZED = "Cube BS Polarizing"

THIN_LENS_TYPE = 0,
D_SHAPED_TYPE = 1,
ODD_ASPHERE_TYPE = 2,
CONIC_TYPE = 3,
SPHERICAL_TYPE = 4,
CYLINDRICAL_TYPE = 5,
ELLIPSOIDAL_TYPE = 6,
PARABOLIC_TYPE = 7,
ASPHERIC_TYPE = 8,
HALF_BALL_LENS_TYPE = 9,
BALL_LENS_TYPE = 10,
OFF_AXIS_PARABOLIC_TYPE = 11,
ROOF_PRISM_MIRROR_TYPE = 12,
BS_CUBE_TYPE = 13,
RIGHT_ANGLE_PRISM_TYPE = 14,
EQUAL_PRISM_TYPE = 15,
WEDGE_TYPE = 16,
PENTA_PRISM_TYPE = 17,
DISPERSION_PRISM_TYPE = 18,
PELLIN_BROCA_PRISM_TYPE = 19,
DOVE_PRISM_TYPE = 20,
AXICON_PRISM_TYPE = 21,
RHOMBIC_PRISM_TYPE = 22,
XY_POLYNOMIAL_TYPE = 23,
ODD_POLYNOMIAL_TYPE = 24,
QCON_ASPHERE_TYPE = 25,
Y_TOROID_TYPE = 26,
X_TOROID_TYPE = 27,
GRATING_TYPE = 28,
MULTIPLET_TYPE = 29,
PARAXIAL_LENS_TYPE = 30,
ZERNIKE_TYPE = 31,
MASK_TYPE = 32,
APERTURE_TYPE = 33
